@@ -7,6 +7,7 @@
 
 #include "logging.h"
 #include "pytypes.h"
+#include "structure.h"
 #include "version.h"
 
 namespace pystack {
@@ -88,17 +89,15 @@ TupleObject::TupleObject(
 {
     d_manager = manager;
 
-    PyTupleObject tuple;
-    manager->copyObjectFromProcess(addr, &tuple);
-
-    ssize_t num_items = tuple.ob_base.ob_size;
+    Structure<py_tuple_v> tuple(manager, addr);
+    ssize_t num_items = tuple.getField(&py_tuple_v::o_ob_size);
     if (num_items == 0) {
         LOG(DEBUG) << std::hex << std::showbase << "There are no elements in this tuple";
         return;
     }
     d_items.resize(num_items);
     manager->copyMemoryFromProcess(
-            addr + offsetof(PyTupleObject, ob_item),
+            tuple.getFieldRemoteAddress(&py_tuple_v::o_ob_item),
             num_items * sizeof(PyObject*),
             d_items.data());
 }
@@ -120,17 +119,15 @@ ListObject::ListObject(const std::shared_ptr<const AbstractProcessManager>& mana
 {
     d_manager = manager;
 
-    PyListObject list;
-    manager->copyObjectFromProcess(addr, &list);
-
-    ssize_t num_items = list.ob_base.ob_size;
+    Structure<py_list_v> list(manager, addr);
+    ssize_t num_items = list.getField(&py_list_v::o_ob_size);
     if (num_items == 0) {
         LOG(DEBUG) << std::hex << std::showbase << "There are no elements in this list";
         return;
     }
     d_items.resize(num_items);
     manager->copyMemoryFromProcess(
-            (remote_addr_t)list.ob_item,
+            (remote_addr_t)list.getField(&py_list_v::o_ob_item),
             num_items * sizeof(PyObject*),
             d_items.data());
 }
@@ -160,9 +157,20 @@ LongObject::LongObject(
     constexpr unsigned int shift = 15;
 #endif
 
-    _PyLongObject longobj;
-    manager->copyObjectFromProcess(addr, &longobj);
-    ssize_t size = longobj.ob_base.ob_size;
+    Structure<py_long_v> longobj(manager, addr);
+    ssize_t size;
+    bool negative;
+
+    Py_ssize_t ob_size = longobj.getField(&py_long_v::o_ob_size);
+    if (manager->versionIsAtLeast(3, 12)) {
+        auto lv_tag = *reinterpret_cast<uintptr_t*>(&ob_size);
+        negative = (lv_tag & 3) == 2;
+        size = lv_tag >> 3;
+    } else {
+        negative = ob_size < 0;
+        size = std::abs(ob_size);
+    }
+
     if (size == 0) {
         d_value = 0;
         return;
@@ -186,12 +194,12 @@ LongObject::LongObject(
      */
 
     std::vector<digit> digits;
-    digits.resize(std::abs(size));
+    digits.resize(size);
     manager->copyMemoryFromProcess(
-            addr + offsetof(_PyLongObject, ob_digit),
-            sizeof(digit) * std::abs(size),
+            longobj.getFieldRemoteAddress(&py_long_v::o_ob_digit),
+            sizeof(digit) * size,
             digits.data());
-    for (ssize_t i = 0; i < std::abs(size); ++i) {
+    for (ssize_t i = 0; i < size; ++i) {
         long long factor;
         if (__builtin_mul_overflow(digits[i], (1Lu << (ssize_t)(shift * i)), &factor)) {
             d_overflowed = true;
@@ -203,7 +211,9 @@ LongObject::LongObject(
         }
     }
 
-    d_value = size < 0 ? -1 * d_value : d_value;
+    if (negative) {
+        d_value = -1 * d_value;
+    }
 }
 
 std::string
@@ -236,26 +246,24 @@ LongObject::Overflowed() const
 void
 getDictEntries(
         const std::shared_ptr<const AbstractProcessManager>& manager,
-        const Python3::PyDictObject& dict,
+        Structure<py_dict_v>& dict,
         ssize_t& num_items,
         std::vector<Python3::PyDictKeyEntry>& valid_entries)
 {
-    auto keys_addr = reinterpret_cast<remote_addr_t>(dict.ma_keys);
-    assert(manager->majorVersion() == 3);
+    assert(manager->versionIsAtLeast(3, 0));
+    remote_addr_t keys_addr = dict.getField(&py_dict_v::o_ma_keys);
     ssize_t dk_size = 0;
     int dk_kind = 0;
 
-    if (manager->minorVersion() <= 10) {
-        Python3_3::PyDictKeysObject keys;
-        manager->copyObjectFromProcess(keys_addr, &keys);
-        num_items = keys.dk_nentries;
-        dk_size = keys.dk_size;
-    } else {
-        Python3_11::PyDictKeysObject keys;
-        manager->copyObjectFromProcess(keys_addr, &keys);
-        num_items = keys.dk_nentries;
-        dk_size = 1L << keys.dk_log2_size;
-        dk_kind = keys.dk_kind;
+    Structure<py_dictkeys_v> keys(manager, keys_addr);
+    num_items = keys.getField(&py_dictkeys_v::o_dk_nentries);
+    dk_size = keys.getField(&py_dictkeys_v::o_dk_size);
+
+    if (manager->versionIsAtLeast(3, 11)) {
+        // We're reusing the o_dk_size offset for dk_log2_size. Fix up the value.
+        dk_size = 1L << dk_size;
+        // Added in 3.11
+        dk_kind = keys.getField(&py_dictkeys_v::o_dk_kind);
     }
     if (num_items == 0) {
         LOG(DEBUG) << std::hex << std::showbase << "There are no elements in this dict";
@@ -280,14 +288,8 @@ getDictEntries(
         offset = 8 * dk_size;
     }
 
-    offset_t dk_indices_offset = 0;
-    if (manager->minorVersion() <= 10) {
-        dk_indices_offset = offsetof(Python3_3::PyDictKeysObject, dk_indices);
-    } else {
-        dk_indices_offset = offsetof(Python3_11::PyDictKeysObject, dk_indices);
-    }
-
-    remote_addr_t entries_addr = keys_addr + dk_indices_offset + offset;
+    offset_t dk_indices_addr = keys.getFieldRemoteAddress(&py_dictkeys_v::o_dk_indices);
+    remote_addr_t entries_addr = dk_indices_addr + offset;
 
     std::vector<Python3::PyDictKeyEntry> raw_entries;
     raw_entries.resize(num_items);
@@ -344,12 +346,10 @@ DictObject::DictObject(std::shared_ptr<const AbstractProcessManager> manager, re
     // For now, the layout that we use here only allows us to get Python3.6+ dictionaries
     // as dictionaries before that have much more variability and are much harder to get.
 
-    if (d_manager->majorVersion() > 2) {
-        if (d_manager->minorVersion() < 6) {
-            d_invalid = true;
-            return;
-        }
+    if (d_manager->versionIsAtLeast(3, 6)) {
         loadFromPython3(addr);
+    } else if (d_manager->versionIsAtLeast(3, 0)) {
+        d_invalid = true;
     } else {
         loadFromPython2(addr);
     }
@@ -358,8 +358,7 @@ DictObject::DictObject(std::shared_ptr<const AbstractProcessManager> manager, re
 void
 DictObject::loadFromPython3(remote_addr_t addr)
 {
-    Python3::PyDictObject dict;
-    d_manager->copyObjectFromProcess(addr, &dict);
+    Structure<py_dict_v> dict(d_manager, addr);
 
     ssize_t num_items;
     std::vector<Python3::PyDictKeyEntry> valid_entries;
@@ -388,11 +387,13 @@ DictObject::loadFromPython3(remote_addr_t addr)
      *          All dicts sharing same key must have same insertion order.
      */
 
-    auto values_addr = (remote_addr_t)dict.ma_values;
+    remote_addr_t dictvalues_addr = dict.getField(&py_dict_v::o_ma_values);
+    Structure<py_dictvalues_v> dictvalues(d_manager, dictvalues_addr);
 
     // Get the values in one copy if we are dealing with a split-table dictionary
-    if (values_addr != 0) {
+    if (dictvalues_addr != 0) {
         d_values.resize(num_items);
+        auto values_addr = dictvalues.getFieldRemoteAddress(&py_dictvalues_v::o_values);
         d_manager->copyMemoryFromProcess(values_addr, num_items * sizeof(PyObject*), d_values.data());
     } else {
         std::transform(
@@ -487,7 +488,7 @@ std::string
 GenericObject::toString(ssize_t max_size) const
 {
     std::stringstream os;
-    os << "<" << std::hex << std::showbase << d_classname << " at " << d_addr << ">";
+    os << "<" << std::hex << d_classname << " at 0x" << d_addr << ">";
     std::string object_str = os.str();
     return limitOutput(object_str, max_size);
 }
@@ -505,20 +506,36 @@ NoneObject::toString([[maybe_unused]] ssize_t max_size) const
 
 Object::Object(const std::shared_ptr<const AbstractProcessManager>& manager, remote_addr_t addr)
 : d_addr(addr)
+, d_type_addr(0)
+, d_classname()
+, d_flags()
 , d_manager(manager)
 {
     LOG(DEBUG) << std::hex << std::showbase << "Copying PyObject data from address " << addr;
 
-    PyObject obj;
-    manager->copyObjectFromProcess(d_addr, &obj);
+    Structure<py_object_v> obj(manager, addr);
+    try {
+        obj.copyFromRemote();
+    } catch (RemoteMemCopyError& ex) {
+        LOG(WARNING) << std::hex << std::showbase << "Failed to read PyObject data from address "
+                     << d_addr;
+        d_classname = "invalid object";
+        return;
+    }
 
-    PyTypeObject cls;
-    LOG(DEBUG) << std::hex << std::showbase << "Copying typeobject from address " << obj.ob_type;
-    manager->copyMemoryFromProcess((remote_addr_t)obj.ob_type, manager->offsets().py_type.size, &cls);
+    d_type_addr = obj.getField(&py_object_v::o_ob_type);
+    LOG(DEBUG) << std::hex << std::showbase << "Copying typeobject from address " << d_type_addr;
+    Structure<py_type_v> cls(manager, d_type_addr);
+    try {
+        d_flags = cls.getField(&py_type_v::o_tp_flags);
+    } catch (RemoteMemCopyError& ex) {
+        LOG(WARNING) << std::hex << std::showbase << "Failed to read typeobject from address "
+                     << d_type_addr;
+        d_classname = "invalid object";
+        return;
+    }
 
-    d_flags = manager->versionedTypeField<unsigned long, &py_type_v::o_tp_flags>(cls);
-
-    remote_addr_t name_addr = manager->versionedTypeField<remote_addr_t, &py_type_v::o_tp_name>(cls);
+    remote_addr_t name_addr = cls.getField(&py_type_v::o_tp_name);
     try {
         d_classname = manager->getCStringFromAddress(name_addr);
     } catch (RemoteMemCopyError& ex) {
@@ -536,6 +553,12 @@ Object::hasFlags(unsigned long flags) const
     return flags & d_flags;
 }
 
+remote_addr_t
+Object::typeAddr() const
+{
+    return d_type_addr;
+}
+
 // Helpers for making overloaded lambdas in the variant visitor in Object::toString
 template<class... Ts>
 struct overloaded : Ts...
@@ -554,24 +577,33 @@ Object::toString(ssize_t max_size) const
     if (max_size <= 5) {
         return ELLIPSIS;
     }
-    std::stringstream os;
-    std::visit(
-            overloaded{
-                    [&](const auto& arg) { os << arg.toString(max_size); },
-                    [&](const bool arg) { os << arg; },
-                    [&](const long arg) { os << arg; },
-                    [&](const double arg) { os << arg; },
-                    [&](const std::string& arg) {
-                        std::string truncated = arg;
-                        if (static_cast<size_t>(max_size) < arg.size()) {
-                            truncated = arg.substr(0, max_size - 3) + "...";
-                        }
-                        os << truncated;
-                    },
-            },
-            toConcreteObject());
+    try {
+        std::stringstream os;
+        std::visit(
+                overloaded{
+                        [&](const auto& arg) { os << arg.toString(max_size); },
+                        [&](const bool arg) { os << arg; },
+                        [&](const long arg) { os << arg; },
+                        [&](const double arg) { os << arg; },
+                        [&](const std::string& arg) {
+                            std::string truncated = arg;
+                            if (static_cast<size_t>(max_size) < arg.size()) {
+                                truncated = arg.substr(0, max_size - 3) + "...";
+                            }
+                            os << truncated;
+                        },
+                },
+                toConcreteObject());
 
-    return os.str();
+        return os.str();
+    } catch (RemoteMemCopyError& ex) {
+        LOG(WARNING) << std::hex << std::showbase << "Failed to create a repr for object of type "
+                     << d_classname << " at address " << d_addr;
+
+        std::stringstream os;
+        os << std::hex << std::showbase << "<" << d_classname << " object at " << d_addr << ">";
+        return os.str();
+    }
 }
 
 long
@@ -585,9 +617,8 @@ Object::toInteger() const
 double
 Object::toFloat() const
 {
-    PyFloatObject the_float;
-    d_manager->copyObjectFromProcess(d_addr, &the_float);
-    return the_float.ob_fval;
+    Structure<py_float_v> the_float(d_manager, d_addr);
+    return the_float.getField(&py_float_v::o_ob_fval);
 }
 
 bool
@@ -617,9 +648,8 @@ Object::objectType() const
 
     const long subclass_flags = d_flags & subclass_mask;
 
-
     if (subclass_flags == Pystack_TPFLAGS_BYTES_SUBCLASS) {
-        return d_manager->majorVersion() > 2 ? ObjectType::BYTES : ObjectType::STRING;
+        return d_manager->versionIsAtLeast(3, 0) ? ObjectType::BYTES : ObjectType::STRING;
     } else if (subclass_flags == Pystack_TPFLAGS_UNICODE_SUBCLASS) {
         return ObjectType::STRING;
     } else if (subclass_flags == Pystack_TPFLAGS_INT_SUBCLASS) {
@@ -642,6 +672,8 @@ Object::objectType() const
         return ObjectType::FLOAT;
     } else if (d_classname == "NoneType") {
         return ObjectType::NONE;
+    } else if (d_classname == "code") {
+        return ObjectType::CODE;
     }
     return ObjectType::OTHER;
 }
@@ -652,12 +684,10 @@ Object::toConcreteObject() const
     try {
         switch (objectType()) {
             case Object::ObjectType::STRING:
-                if (d_manager->majorVersion() < 3) {
-                    return normalizeBytesObjectRepresentation(
-                            d_manager->getStringFromAddress(d_addr),
-                            "");
+                if (d_manager->versionIsAtLeast(3, 0)) {
+                    return '"' + d_manager->getStringFromAddress(d_addr) + '"';
                 }
-                return '"' + d_manager->getStringFromAddress(d_addr) + '"';
+                return normalizeBytesObjectRepresentation(d_manager->getStringFromAddress(d_addr), "");
             case Object::ObjectType::BYTES:
                 return normalizeBytesObjectRepresentation(d_manager->getBytesFromAddress(d_addr));
             case Object::ObjectType::NONE:
@@ -678,6 +708,7 @@ Object::toConcreteObject() const
                 return ListObject(d_manager, d_addr);
             case Object::ObjectType::DICT:
                 return DictObject(d_manager, d_addr);
+            case Object::ObjectType::CODE:
             case Object::ObjectType::OTHER:
                 return GenericObject(d_addr, d_classname);
         }
@@ -688,9 +719,9 @@ Object::toConcreteObject() const
 }
 
 std::string
-Object::guessClassName(PyTypeObject& type) const
+Object::guessClassName(Structure<py_type_v>& type) const
 {
-    remote_addr_t tp_repr = d_manager->versionedTypeField<remote_addr_t, &py_type_v::o_tp_repr>(type);
+    remote_addr_t tp_repr = type.getField(&py_type_v::o_tp_repr);
     if (tp_repr == d_manager->findSymbol("float_repr")) {
         return "float";
     }
@@ -700,6 +731,9 @@ Object::guessClassName(PyTypeObject& type) const
     if (tp_repr == d_manager->findSymbol("bool_repr")) {
         return "bool";
     }
+    if (tp_repr == d_manager->findSymbol("code_repr")) {
+        return "PyCodeObject";
+    }
     return "???";
 }
 
